@@ -7,7 +7,7 @@
  *  UCRL-CODE-2003-006.
  *  
  *  This file is part of Scrub, a program for erasing disks.
- *  For details, see <http://www.llnl.gov/linux/scrub/>.
+ *  For details, see https://code.google.com/p/diskscrub/
  *  
  *  Scrub is free software; you can redistribute it and/or modify it under
  *  the terms of the GNU General Public License as published by the Free
@@ -24,7 +24,7 @@
  *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 \*****************************************************************************/
 
-off_t getsize(char *path);
+int getsize(char *path, off_t *sizep);
 off_t str2size(char *str);
 int str2int(char *str);
 void size2str(char *str, int len, off_t size);
